require 'net/http'
require 'json'
require 'base64'
require 'date'

module Arko
  extend self
  
  module MainUIService
    extend self
  
    module ArkoConfig
      extend self

      def self.version
        '2.1.0'
      end

      def self.ui_url
        # 'C:\Users\seldonp\arko\arko.sketchup\a_arko\index.html' # local
        File.join(__dir__,'index.html') # new prod
      end
    end

    @current_version = ArkoConfig.version
    @model = Sketchup.active_model
    @dialog = nil
    @selection_observer = nil
    @app_observer = nil,
    @token = ''

    @main_ui_options = {
      dialog_title: "ArkoAI | #{@current_version}",
      style: UI::HtmlDialog::STYLE_DIALOG,
      width: 875,
      height: 640,
      scrollable: false,
      min_width: 875,
      min_height: 640
    }

    def self.show_dialog
      ArkoFile.create_all_directories

      @dialog ||= create_dialog

      $arko_dialog = @dialog
      dialog_src

      register_endpoints
      register_observers

      @dialog.show
    end

    def self.create_dialog
      @dialog = UI::HtmlDialog.new(@main_ui_options)
      @dialog.center
      $arko_dialog = @dialog
      @dialog
    end

    def self.dialog_src
      # url = 'http://localhost:9000'
      url = ArkoConfig.ui_url
      @dialog.set_file(url)
    end

    def self.register_endpoints
      @dialog.add_action_callback('checkVersion') do |_|
        ArkoFile.save_install_datetime
        @dialog.execute_script("arko.setInstallDateTime('#{ArkoFile.read_install_datetime}')")
        @dialog.execute_script("arko.setVersion('#{@current_version}')")
        @dialog.execute_script("arko.processGAEvent('arko_version', '#{@current_version}')")
      end

      @dialog.add_action_callback('save_user_credentials') do |_, email, password|
        ArkoFile.save_user_credentials(email, password)
      end

      @dialog.add_action_callback('logoutClick') do |_|
        puts 'logout'
        ArkoFile.delete_credentials
        @dialog.close
      end

      @dialog.add_action_callback('tryGetCredentials') do |_|
        content = ArkoFile.read_user_credentials

        if content.nil?
          puts 'no user credentials found - starting sign in process'
        else
          credentials = JSON.parse(content)
          email = credentials['Email']
          password = credentials['Password']

          @dialog.execute_script("arko.silentSignIn('#{email}', '#{password}')")
        end
      rescue
        puts 'issue with tryGetCredentials'
        @dialog.execute_script("arko.processGAEvent('ruby_tryGetCredentials_error', '')")
        raise
      end

      @dialog.add_action_callback('requestActiveView') do |_|
        puts 'requestActiveView'
        Camera.save_current_view

        current_active_view_base64 = Camera.current_img_base64
        
        model = Sketchup.active_model
        view = model.active_view

        width, height = Camera.view_dimensions(view, 5)

        @dialog.execute_script("arko.refreshActiveView(#{width}, #{height}, '#{current_active_view_base64}')")
      rescue
        @dialog.execute_script("arko.processGAEvent('ruby_refresh_active_view_error', '')")
        raise
      end
      
      @dialog.add_action_callback('openRenderDirectory') do |_|
        ArkoFile.open_render_dir
      rescue
        @dialog.execute_script("arko.processGAEvent('ruby_open_render_directory_error', '')")
      end

      @dialog.add_action_callback('renderRequest') do |_, prompt, prompt_importance, similarity_to_original, category, seed, negative_prompt, image_quality|
        puts 'getRenderRequest'
        @dialog.execute_script("arko.updateRenderProgress('running')")
        send_render_response(prompt, prompt_importance, similarity_to_original, category, seed, negative_prompt, image_quality)
      rescue => e
        puts 'issue with rendering request'
        @dialog.execute_script("arko.processGAEvent('ruby_render_request_process_error', '')")
        @dialog.execute_script("arko.updateRenderProgress('error')")
        raise
      end

      @dialog.add_action_callback('completeRender') do |_, render_base64|
        puts 'completeRender'
        puts 'saving image'
        ArkoFile.save_image(render_base64)
        @dialog.execute_script("arko.updateRenderProgress('success')")
      rescue => e
        puts 'issue rendering image'
        @dialog.execute_script("arko.processGAEvent('ruby_render_image_error', '')")
        @dialog.execute_script("arko.updateRenderProgress('error')")
        raise
      end
    end

    def self.send_render_response(prompt, prompt_importance, similarity_to_original, category, seed, negative_prompt, image_quality)
      model = Sketchup.active_model
      view = model.active_view

      width, height = Camera.view_dimensions(view, image_quality)

      render_request = JSON.dump({
        'image' => Camera.current_img_base64,
        'prompt' => "#{prompt}, #{SD.prompt_by_category(category)}, #{SD.positive_prompt}",
        'prompt_importance' => prompt_importance.to_i / 3.33,
        'similarity_to_original' => similarity_to_original.to_i / 100.0,
        'width' => width,
        'height' => height,
        'seed' => seed,
        'negative_prompt' => negative_prompt,
        'image_quality' => image_quality
      })

      @dialog.execute_script("arko.renderRequest('#{render_request}')")
    end

    def self.register_observers
      Sketchup.remove_observer @app_observer if @app_observer

      @app_observer = AppObserver.new(@dialog, @model)

      Sketchup.add_observer(@app_observer)
    end

    class AppObserver < Sketchup::AppObserver
      @dialog = nil
      @model = nil

      def initialize(dialog, model)
        @dialog = dialog
        @model = model
        super()
      end

      def onNewModel(model)
        # observe_model(model)
      end

      def onOpenModel(model)
        # observe_model(model)
      end

      def expectsStartupModelNotifications
        true
      end
    end

    module ArkoFile
      extend self

      def self.set_directories
        if Sketchup.platform == :platform_win
          @arko_dir = "#{ENV['ProgramData']}\\ArkoAI"
          @arko_temp_dir = "#{ENV['ProgramData']}\\ArkoAI\\Temp"
          @arko_user_dir = "#{ENV['ProgramData']}\\ArkoAI\\User"
          @arko_render_dir = "#{ENV['ProgramData']}\\ArkoAI\\Sketchup"
        elsif Sketchup.platform == :platform_osx
          @arko_dir = File.expand_path('~/Documents/ArkoAI')
          @arko_temp_dir = File.expand_path('~/Documents/ArkoAI/Temp')
          @arko_user_dir = File.expand_path('~/Documents/ArkoAI/User')
          @arko_render_dir = File.expand_path('~/Documents/ArkoAI/Sketchup')
        end
      end

      def self.create_all_directories
        set_directories
        Dir.mkdir(@arko_dir) unless File.exists?(@arko_dir)
        Dir.mkdir(@arko_temp_dir) unless File.exists?(@arko_temp_dir)
        Dir.mkdir(@arko_user_dir) unless File.exists?(@arko_user_dir)
        Dir.mkdir(@arko_render_dir) unless File.exists?(@arko_render_dir)
      end

      def self.get_temp_dir
        @arko_temp_dir
      end

      def self.open_render_dir
        puts 'open_render_dir'
        puts "explorer #{@arko_render_dir}"
        system("explorer #{@arko_render_dir}")
      end

      def self.save_user_credentials(email, password)
        credentials = JSON.dump({
          'Email': email,
          'Password': password
        })

        File.open("#{@arko_user_dir}/arko.user", 'w+') { |file| file.write(credentials) }
      end

      def self.save_image(imageBase64)
        now = DateTime.now
        file_name = "ArkoAI-#{now.year}#{now.month}#{now.day}#{now.hour}#{now.min}#{now.sec}-render.png"

        #{Base64.strict_encode64(File.open(@current_view_path, "rb").read)}

        File.open("#{@arko_render_dir}/#{file_name}", 'wb+') { |file|
          file.write(Base64.decode64(imageBase64.split(',')[1]))
        }
      end

      def self.read_user_credentials
        return unless File.exists?("#{@arko_user_dir}/arko.user")

        File.read("#{@arko_user_dir}/arko.user")
      end

      def self.read_install_datetime
        return unless File.exists?("#{@arko_user_dir}/initialize.user")

        File.read("#{@arko_user_dir}/initialize.user")
      end

      def self.save_install_datetime
        return if File.exists?("#{@arko_user_dir}/initialize.user")

        now = DateTime.now
        install_datetime = "#{now.year}#{now.month}#{now.day}#{now.hour}#{now.min}#{now.sec}"
        File.open("#{@arko_user_dir}/initialize.user", 'w+') { |file| file.write(install_datetime) }
      end

      def self.delete_credentials
        credentials_file_name = "#{@arko_user_dir}/arko.user"
        File.delete(credentials_file_name) if File.exists?(credentials_file_name)
      end
    end

    module Camera
      extend self

      def self.get_current_view_path
        "#{ArkoFile.get_temp_dir}\\arko-current-view.png"
      end

      def self.current_img_base64
        return "data:image/png;base64,#{Base64.strict_encode64(File.open(get_current_view_path, "rb").read)}"
      end

      def self.save_current_view
        model = Sketchup.active_model
        view = model.active_view

        width, height = view_dimensions(view, 5)

        options = {
          :filename => get_current_view_path,
          :width => width,
          :height => height,
        }

        puts "saving current view to: #{get_current_view_path}"
        view.write_image(options)
      rescue
        puts 'Issue capturing current view'
        raise
      end

      def round_up(n, increment)
        increment * (( n + increment - 1) / increment)
      end

      def view_dimensions(view, image_quality)
        width = view.vpwidth
        height = view.vpheight

        # puts 'original wh'
        # puts width
        # puts height

        # max_size = 512
        # if width < height
        #   ratio = height / width.to_f
        #   width = max_size
        #   height = (max_size * ratio).ceil
        # else
        #   ratio = width / height.to_f
        #   height = max_size
        #   width = (max_size * ratio).ceil
        # end

        # TODO set height / width by image_quality
        max_size_2 = 1344
        max_size_1 = 786

        # lesser of width / height becomes max_size_1
        # greater size determined by aspect ratio or max_size_2, whichever is less
        # max size originally 786 x 1344
        if width < height
          ratio = height / width.to_f
          width = max_size_1
          height_by_aspect_ratio = (max_size_1 * ratio).ceil
          height = height_by_aspect_ratio <= max_size_2 ? height_by_aspect_ratio : max_size_2
        else
          ratio = width / height.to_f
          height = max_size_1
          height_by_aspect_ratio = (max_size_1 * ratio).ceil
          width = height_by_aspect_ratio <= max_size_2 ? height_by_aspect_ratio : max_size_2
        end

        # puts 'new wh'
        # puts width
        # puts height
        # puts image_quality_percent(image_quality)
        width = ((image_quality_percent(image_quality) * width) / 100.0).ceil
        height = ((image_quality_percent(image_quality) * height) / 100.0).ceil

        # puts 'new wh'
        # puts width
        # puts height

        w, h = image_as_multiple_of_8(width, height)
        
        # puts 'final wh'
        # puts w
        # puts height

        return w, h
      end

      def image_as_multiple_of_8(width, height)
        if width > height
          larger_side = width
          smaller_side = height

          original_ratio = smaller_side.to_f / larger_side
          larger_side_8 = round_up(larger_side, 8)
          smaller_side_8 = round_up((larger_side_8 * original_ratio).round, 8)

          height = smaller_side_8
          width = larger_side_8
        else
          larger_side = height
          smaller_side = width

          original_ratio = smaller_side.to_f / larger_side
          larger_side_8 = round_up(larger_side, 8)
          smaller_side_8 = round_up((larger_side_8 * original_ratio).round, 8)

          height = larger_side_8
          width = smaller_side_8
        end

        return width, height
      end

      def image_quality_percent(image_quality)
        if image_quality == 5
          return 100
        elsif image_quality == 4
          return 90
        elsif image_quality == 3
          return 80
        elsif image_quality == 2
          return 70
        elsif image_quality == 1
          return 60
        else
          return 100
        end
      end
    end

    module SD
      extend self
      @categories = {
        'Generic/None' => '',
        'Generic/Nature' => '((Nature)), forest, plants', # start pro categories
        'Generic/Snow' => 'Snow", "((Heavy Snow))',
        'Nature' => '(Nature), Forest, vegetation, trees, landscape, rocks, river, mountains',
        'Architecture' => '(Architecture), Design, Building Design',
        'Architecture/Ancient Roman' => '(architecture), Mid century modern, (Ancient Roman), Design, Building Design',
        'Architecture/Brutalist' => '(architecture), Mid century modern, (Brutalist style), Design, Building Design',
        'Architecture/Chinese' => '(architecture), Mid century modern, (Chinese style), Design, Building Design',
        'Architecture/Gothic' => '(architecture), Mid century modern, (Gothic style), Design, Building Design',
        'Architecture/Industrial' => '(architecture), Mid century modern, (Industrial style), Design, Building Design',
        'Architecture/Minimalist' => '(architecture), Mid century modern, (Minimalist style), Design, Building Design',
        'Architecture/Modern' => '(architecture), Mid century modern, Design, Building Design',
        'Architecture/Business' => '(architecture), Mid century modern, (retail), office building, Design, Building Design',
        'Architecture/Educational' => '(architecture), Mid century modern, (Educational), Design, Building Design',
        'Architecture/Institutional' => '(architecture), Mid century modern, (Institutional), Design, Building Design',
        'Architecture/Mixed Used' => '(architecture), Mid century modern, (Mixed Used Building),Design, Building Design',
        'Architecture/Residential' => '(architecture), Mid century modern, (residential), Design, Building Design',
        'Architecture/Skyscraper' => '(architecture), Mid century modern, (Skyscraper), Design, Building Design',
        'Architecture/Storage' => '(architecture), Mid century modern, (Storage architecture), Design, Building Design',
        'Architecture/Lego' => '(architecture), (Lego architecture), Mid century modern, Design, Building Design',
        'Interior Design/Chinese' => '(Interior Design), Mid century modern, Chinese interior design',
        'Interior Design/Classical' => '(Interior Design), Mid century modern, Classical interior design',
        'Interior Design/Industrial' => '(Interior Design), Mid century modern, Industrial interior design',
        'Interior Design/Minimalist' => '(Interior Design), Mid century modern, Minimalist interior design',
        'Interior Design/Educational' => '(Interior Design), Mid century modern, Educational interior design',
        'Interior Design/Hospitality' => '(Interior Design), Mid century modern, Hospitality interior design',
        'Interior Design/Institutional' => '(Interior Design), Mid century modern, Institutional interior design',
        'Interior Design/Residential' => '(Interior Design), Mid century modern, Residential interior design',
        'Interior Design/Restaurant' => '(Interior Design), Mid century modern, Restaurant interior design',
        'Interior Design/Retail' => '(Interior Design), Mid century modern, Retail interior design',
        'Interior Design/Workplace' => '(Interior Design), Mid century modern, Workplace interior design',
        'Landscape Architecture' => '(Landscape Architecture), vegetation, trees, landscape, rocks',
        'Landscape Architecture/Minimalistic' => 'minimalistic, (Landscape Architecture), vegetation, trees, landscape, rocks',
        'Landscape Architecture/Modern' => 'modern, (Landscape Architecture), vegetation, trees, landscape, rocks',
        'Landscape Architecture/Parks & Recreation' => 'park with trees, (Landscape Architecture), vegetation, trees, landscape, rocks',
        'Landscape Architecture/Urban Design' => 'urban design, (Landscape Architecture), vegetation, trees, landscape, rocks',
        'Landscape Architecture/Garden' => 'garden, (Landscape Architecture), vegetation, trees, landscape, rocks',
        'Construction' => '(Construction), engineering',
        'Interior Design' => '(Interior Design), building interior, Design',
        'Industrial Design' => '(Industrial Design), Design'
      }
      @positive_prompt = "sharp focus, wide-angle, Fujifilm XT3, Canon Eos 5D, tone mapping"
      @negative_prompt = "(((text))), blurry, watermark, words, captions, lowres, error, cropped, worst quality, low quality, jpeg artifacts, ugly, duplicate, out of frame, blurry"

      def self.prompt_by_category(category)
        @categories[category]
      end

      def self.positive_prompt
        @positive_prompt
      end

      def self.negative_prompt
        @negative_prompt
      end
    end

    file_loaded(__FILE__)
  end
end
